@@ -29,6 +29,7 @@ from bson import (CodecOptions,
                   _dict_to_bson,
                   _make_c_string)
 from bson.codec_options import DEFAULT_CODEC_OPTIONS
+from bson.raw_bson import _inflate_bson, DEFAULT_RAW_BSON_OPTIONS
 from bson.py3compat import b, StringIO
 from bson.son import SON
 
@@ -47,6 +48,7 @@ from pymongo.errors import (ConfigurationError,
                             ProtocolError)
 from pymongo.read_concern import DEFAULT_READ_CONCERN
 from pymongo.read_preferences import ReadPreference
+from pymongo.write_concern import WriteConcern
 
 
 MAX_INT32 = 2147483647
@@ -231,6 +233,10 @@ class _Query(object):
                  'batch_size', 'name', 'read_concern', 'collation',
                  'session', 'client', '_as_command')
 
+    # For compatibility with the _GetMore class.
+    exhaust_mgr = None
+    cursor_id = None
+
     def __init__(self, flags, db, coll, ntoskip, spec, fields,
                  codec_options, read_preference, limit,
                  batch_size, read_concern, collation, session, client):
@@ -251,6 +257,9 @@ class _Query(object):
         self.name = 'find'
         self._as_command = None
 
+    def namespace(self):
+        return _UJOIN % (self.db, self.coll)
+
     def use_command(self, sock_info, exhaust):
         use_find_cmd = False
         if sock_info.max_wire_version >= 4:
@@ -289,7 +298,7 @@ class _Query(object):
             cmd = SON([('explain', cmd)])
         session = self.session
         if session:
-            session._apply_to(cmd, False, self.read_preference, sock_info)
+            session._apply_to(cmd, False, self.read_preference)
             # Explain does not support readConcern.
             if (not explain and session.options.causal_consistency
                     and session.operation_time is not None
@@ -298,6 +307,12 @@ class _Query(object):
                     'readConcern', {})[
                     'afterClusterTime'] = session.operation_time
         sock_info.send_cluster_time(cmd, session, self.client)
+        # Support auto encryption
+        client = self.client
+        if (client._encrypter and
+                not client._encrypter._bypass_auto_encryption):
+            cmd = client._encrypter.encrypt(
+                self.db, cmd, False, self.codec_options)
         self._as_command = cmd, self.db
         return self._as_command
 
@@ -309,7 +324,7 @@ class _Query(object):
         else:
             flags = self.flags
 
-        ns = _UJOIN % (self.db, self.coll)
+        ns = self.namespace()
         spec = self.spec
 
         if use_cmd:
@@ -347,12 +362,13 @@ class _GetMore(object):
 
     __slots__ = ('db', 'coll', 'ntoreturn', 'cursor_id', 'max_await_time_ms',
                  'codec_options', 'read_preference', 'session', 'client',
-                 '_as_command')
+                 'exhaust_mgr', '_as_command')
 
     name = 'getMore'
 
     def __init__(self, db, coll, ntoreturn, cursor_id, codec_options,
-                 read_preference, session, client, max_await_time_ms=None):
+                 read_preference, session, client, max_await_time_ms,
+                 exhaust_mgr):
         self.db = db
         self.coll = coll
         self.ntoreturn = ntoreturn
@@ -362,8 +378,12 @@ class _GetMore(object):
         self.session = session
         self.client = client
         self.max_await_time_ms = max_await_time_ms
+        self.exhaust_mgr = exhaust_mgr
         self._as_command = None
 
+    def namespace(self):
+        return _UJOIN % (self.db, self.coll)
+
     def use_command(self, sock_info, exhaust):
         sock_info.validate_session(self.client, self.session)
         return sock_info.max_wire_version >= 4 and not exhaust
@@ -379,15 +399,21 @@ class _GetMore(object):
                                     self.max_await_time_ms)
 
         if self.session:
-            self.session._apply_to(cmd, False, self.read_preference, sock_info)
+            self.session._apply_to(cmd, False, self.read_preference)
         sock_info.send_cluster_time(cmd, self.session, self.client)
+        # Support auto encryption
+        client = self.client
+        if (client._encrypter and
+                not client._encrypter._bypass_auto_encryption):
+            cmd = client._encrypter.encrypt(
+                self.db, cmd, False, self.codec_options)
         self._as_command = cmd, self.db
         return self._as_command
 
     def get_message(self, dummy0, sock_info, use_cmd=False):
         """Get a getmore message."""
 
-        ns = _UJOIN % (self.db, self.coll)
+        ns = self.namespace()
         ctx = sock_info.compression_context
 
         if use_cmd:
@@ -838,10 +864,10 @@ class _BulkWriteContext(object):
 
     __slots__ = ('db_name', 'command', 'sock_info', 'op_id',
                  'name', 'field', 'publish', 'start_time', 'listeners',
-                 'session', 'compress')
+                 'session', 'compress', 'op_type', 'codec')
 
     def __init__(self, database_name, command, sock_info, operation_id,
-                 listeners, session):
+                 listeners, session, op_type, codec):
         self.db_name = database_name
         self.command = command
         self.sock_info = sock_info
@@ -853,6 +879,38 @@ class _BulkWriteContext(object):
         self.start_time = datetime.datetime.now() if self.publish else None
         self.session = session
         self.compress = True if sock_info.compression_context else False
+        self.op_type = op_type
+        self.codec = codec
+
+    def _batch_command(self, docs):
+        namespace = self.db_name + '.$cmd'
+        request_id, msg, to_send = _do_bulk_write_command(
+            namespace, self.op_type, self.command, docs, self.check_keys,
+            self.codec, self)
+        if not to_send:
+            raise InvalidOperation("cannot do an empty bulk write")
+        return request_id, msg, to_send
+
+    def execute(self, docs, client):
+        request_id, msg, to_send = self._batch_command(docs)
+        result = self.write_command(request_id, msg, to_send)
+        client._process_response(result, self.session)
+        return result, to_send
+
+    def execute_unack(self, docs, client):
+        request_id, msg, to_send = self._batch_command(docs)
+        # Though this isn't strictly a "legacy" write, the helper
+        # handles publishing commands and sending our message
+        # without receiving a result. Send 0 for max_doc_size
+        # to disable size checking. Size checking is handled while
+        # the documents are encoded to BSON.
+        self.legacy_write(request_id, msg, 0, False, to_send)
+        return to_send
+
+    @property
+    def check_keys(self):
+        """Should we check keys for this operation type?"""
+        return self.op_type == _INSERT
 
     @property
     def max_bson_size(self):
@@ -951,6 +1009,54 @@ class _BulkWriteContext(object):
             request_id, self.sock_info.address, self.op_id)
 
 
+# 2MiB
+_MAX_ENC_BSON_SIZE = 2 * (1024 * 1024)
+# 6MB
+_MAX_ENC_MESSAGE_SIZE = 6 * (1000 * 1000)
+
+
+class _EncryptedBulkWriteContext(_BulkWriteContext):
+    __slots__ = ()
+
+    def _batch_command(self, docs):
+        namespace = self.db_name + '.$cmd'
+        msg, to_send = _encode_batched_write_command(
+            namespace, self.op_type, self.command, docs, self.check_keys,
+            self.codec, self)
+        if not to_send:
+            raise InvalidOperation("cannot do an empty bulk write")
+
+        # Chop off the OP_QUERY header to get a properly batched write command.
+        cmd_start = msg.index(b"\x00", 4) + 9
+        cmd = _inflate_bson(memoryview(msg)[cmd_start:],
+                            DEFAULT_RAW_BSON_OPTIONS)
+        return cmd, to_send
+
+    def execute(self, docs, client):
+        cmd, to_send = self._batch_command(docs)
+        result = self.sock_info.command(
+            self.db_name, cmd, codec_options=_UNICODE_REPLACE_CODEC_OPTIONS,
+            session=self.session, client=client)
+        return result, to_send
+
+    def execute_unack(self, docs, client):
+        cmd, to_send = self._batch_command(docs)
+        self.sock_info.command(
+            self.db_name, cmd, write_concern=WriteConcern(w=0),
+            session=self.session, client=client)
+        return to_send
+
+    @property
+    def max_bson_size(self):
+        """A proxy for SockInfo.max_bson_size."""
+        return min(self.sock_info.max_bson_size, _MAX_ENC_BSON_SIZE)
+
+    @property
+    def max_message_size(self):
+        """A proxy for SockInfo.max_message_size."""
+        return min(self.sock_info.max_message_size, _MAX_ENC_MESSAGE_SIZE)
+
+
 def _raise_document_too_large(operation, doc_size, max_size):
     """Internal helper for raising DocumentTooLarge."""
     if operation == "insert":
@@ -1315,13 +1421,14 @@ def _batched_write_command_impl(
         value = bson.BSON.encode(doc, check_keys, opts)
         # Is there enough room to add this document? max_cmd_size accounts for
         # the two trailing null bytes.
+        doc_too_large = len(value) > max_cmd_size
         enough_data = (buf.tell() + len(key) + len(value)) >= max_cmd_size
         enough_documents = (idx >= max_write_batch_size)
+        if doc_too_large:
+            write_op = list(_FIELD_MAP.keys())[operation]
+            _raise_document_too_large(
+                write_op, len(value), max_bson_size)
         if enough_data or enough_documents:
-            if not idx:
-                write_op = list(_FIELD_MAP.keys())[operation]
-                _raise_document_too_large(
-                    write_op, len(value), max_bson_size)
             break
         buf.write(_BSONOBJ)
         buf.write(key)
@@ -1427,6 +1534,11 @@ class _OpReply(object):
         assert self.number_returned == 1
         return docs[0]
 
+    def raw_command_response(self):
+        """Return the bytes of the command response."""
+        # This should never be called on _OpReply.
+        raise NotImplementedError
+
     @classmethod
     def unpack(cls, msg):
         """Construct an _OpReply from raw bytes."""
@@ -1473,6 +1585,10 @@ class _OpMsg(object):
         """Unpack a command response."""
         return self.unpack_response()[0]
 
+    def raw_command_response(self):
+        """Return the bytes of the command response."""
+        return self.payload_document
+
     @classmethod
     def unpack(cls, msg):
         """Construct an _OpMsg from raw bytes."""
